import time
import serial
from queue import Queue
import threading
from typing import Optional

from .NetworkInterface import NetworkInterface
from ..Config import CNETWORK, CPACKET
from .Packet import Packet

class Serial(NetworkInterface):
    """
    Source
        https://www.instructables.com/Starting-and-Stopping-Python-Threads-With-Events-i/
    """
    _serial_port: str
    _baud_rate: int
    _timeout: int
    _serial: serial.Serial
    _read_queue: Queue[Packet]
    _write_queue: Queue[Packet]
    _stop_event: threading.Event
    _thread: threading.Thread

    def __init__(self, serial_port: str, baud_rate: int = CNETWORK.BAUDRATE, timeout: int = CNETWORK.TIMEOUT) -> None:
        self._serial_port = serial_port
        self._baud_rate = baud_rate
        self._timeout = timeout
        self._serial = None
        self._read_queue = Queue()
        self._write_queue = Queue()
        self._stop_event = threading.Event()

    def connect(self) -> None:
        self._serial = serial.Serial(self._serial_port, self._baud_rate, timeout=self._timeout)

        self._stop_event.clear()#unset the flag

        self._read_thread = threading.Thread(target=self._read_data)
        self._write_thread = threading.Thread(target=self._write_data)

        self._read_thread.start()
        self._write_thread.start()

    def _read_data(self) -> None:
        while not self._stop_event.is_set():
            
            if self._serial.in_waiting > CPACKET.HEADER_SIZE:
                b = self._serial.read(size=1)
                validByte: bool = Packet.verify_flag(b)
                if not validByte:
                    continue

                headerData = self._serial.read(size=CPACKET.HEADER_SIZE)
                packet = Packet()
                packet.deserialize_header(headerData)

                payload_size: int = packet.get_payload_size()
                if self._serial.in_waiting < payload_size:
                    continue
                    
                payloadData = self._serial.read(payload_size)

                packet.deserialize_payload(payloadData)

                self._read_queue.put(packet)

            time.sleep(CNETWORK.BUSY_WAITING_TIMEOUT)
                

    def _write_data(self) -> None:
        while not self._stop_event.is_set():
            if not self._write_queue.empty():
                packet = self._write_queue.get()
                serialisedData = packet.serialize()

                self._serial.write(serialisedData)
            
            time.sleep(CNETWORK.BUSY_WAITING_TIMEOUT)

    def disconnect(self) -> None:
        self._stop_event.set()

        if self._read_thread.is_alive():
            self._read_thread.join()
        
        if self._write_thread.is_alive():
            self._write_thread.join()
        
        if self._serial:
            self._serial.close()

    def check_connection(self) -> bool:
        return True
    
    def get_protocol_name(self) -> str:
        return "Serial"

    def write(self, data: Packet) -> None:
        self._write_queue.put(data)

    def read(self) -> Optional[Packet]:
        if self._read_queue.qsize() > 0:
            return self._read_queue.get()

        return None


from dataclasses import dataclass, asdict
from typing import Dict, Self

@dataclass
class SensorCapabilities:
    """
    a model class which contains capabilities of sensors.

    the capabilities do not include sensor data, only parameters which influence the record behaviour plus identity to indentify the sensor instance.

    Attributes:
        enable (bool): if a sensor is enabled for the record; default is True because when the sensor is created by the user, it is obvious that he wants to enable it. the sensors on the device are set to disabled per default only being enabled when an according capability is passed.
        sample_rate (int): data pointer recored per second
        data_on_state_change (bool): determines of data should be sent from the device if there is no change from the previous state; False => send all data; True => only when change
        identity (str): the identity of the sensor

    """
    identity: str
    sample_rate: int
    data_on_state_change: bool
    enable: bool = True
    include_timestamp: bool = False
    include_sequence: bool = False
    model_data: str = ""

    def to_dict(self) -> Dict:
        """makes a dictionary from the instance"""
        return asdict(self)
    
    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        """
        makes an instance of the according class using a dictionary.

        not existing values are set to default values. identity must exist as it is an integral part.

        Args:
            data (Dict): dictionary with instance properties

        Returns:
            Self: an instance of the class with the assigned values
        """
        default_data = {
            'enable': True,
            'sample_rate': 10,
            'data_on_state_change': False,
            #identity must be provided
        }
        
        merged_data = {**default_data, **data}
        return cls(**merged_data)


import os, importlib, subprocess, uuid
from ..Models.WiFiProfile import WiFiProfile
from ..Network.NetworkInterface import NetworkProtocols
from ..Utils.Logger import LoggerInstance, LoggerType
from ..Config import CGENERAL, CNETWORK

class ArduinoUploader:

    """
    References:
        https://docs.platformio.org/en/latest/projectconf/sections/env/options/build/build_flags.html
    """

    PACKAGE_NAME = "platformio"
    PIO_CONFIG_FILE = "platformio.ini"

    def __init__(self, arduino_path: str, identity: str, serial_port: str, wireless_mode: NetworkProtocols = NetworkProtocols.BLE, wifi_profile: WiFiProfile = None) -> None:
        self.logger: LoggerType = LoggerInstance.get()
        self.arduino_path = arduino_path #relative path
        self.project_path = os.path.join(os.getcwd(), self.arduino_path)
        self.identity = identity
        self.service_uuid = uuid.uuid4()
        self.characteristic_uuid = uuid.uuid4()
        self.serial_port = serial_port
        self.wireless_mode = wireless_mode
        self.wifi_profile = wifi_profile
    
    def check_prerequisits(self) -> bool:
        if not os.path.exists(self.project_path):
            self.logger.error(f"path to arduino directory does not exist")
            return False
        
        #check if the platformio CLI exists and can be used
        try:
            importlib.import_module(self.PACKAGE_NAME)
                    
        except ImportError as e:
            self.logger.error(e)
            return False
        
        #check if platformio.ini data exists in the current working directory. this is a good indicator for a folder with a arduino build to upload.
        pioini_path = os.path.join(self.project_path, self.PIO_CONFIG_FILE)

        if not os.path.exists(pioini_path):
            self.logger.error(f"pioini_path {pioini_path} does not exist")
            return False

        #check if identity is an ascii string and 4 byte long
        if not self.identity.isascii():
            self.logger.error(f"{self.identity} is not of ASCII type")
            return False
        
        if len(self.identity) != 4:
            self.logger.error(f"{self.identity} must be 4 byte long")
            return False

        return True
    
    def _create_flag(self) -> str:
        flags = [
            f'-DMC_NAME=\\"{self.identity}\\"',
            f'-DSERVICE_UUID=\\"{self.service_uuid}\\"',
            f'-DCHARACTERISTIC_UUID=\\"{self.characteristic_uuid}\\"',
            f'-DPORT=\\"{self.serial_port}\\"',
            f'-DBAUDRATE={CNETWORK.BAUDRATE}',
            f'-DDEBUG_MODE={CGENERAL.DEBUG}'
        ]

        if self.wireless_mode == NetworkProtocols.BLE:
            flags.append('-DWIRELESS_MODE=0')
        else:
            flags.extend([
                '-DWIRELESS_MODE=1',
                f'-DSSID=\\"{self.wifi_profile.ssid}\\"',
                f'-DPASSWORD=\\"{self.wifi_profile.password}\\"',
                f'-DCLIENT_IP=\\"{self.wifi_profile.client_ip}\\"',
                f'-DCLIENT_PORT=\\"{self.wifi_profile.client_port}\\"'
            ])

        return flags
    
    def upload(self) -> bool:

        os.environ['PLATFORMIO_BUILD_FLAGS'] = ' '.join(self._create_flag())

        command = [
            self.PACKAGE_NAME, 'run', '-t', 'upload', '--upload-port', self.serial_port, '--project-dir', self.project_path
        ]
             
        try:
            process = subprocess.run(command, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE, check=True)
            self.logger.error(process.stderr)

            return process.returncode == 0
        except subprocess.CalledProcessError as e:
            self.logger.error(f"upload failed: {e}")
        
        except FileNotFoundError as e:
            self.logger.error(f"command unknown: {e}")

        return False

    
if __name__ == "__main__":
    #upload a build with wifi enabled; ble requires a different build     
    wifi_profile = WiFiProfile(
        wifi_key="WIFI_PROFILE_2",
        ssid="172.168.192.0",
        password="1829746238267463950124",
        client_ip="172.168.192.27",
        client_port=8001
    )

    a = ArduinoUploader("./arduino", "MC01", "/dev/ttyUSB0", NetworkProtocols.WIFI, wifi_profile)
    if a.check_prerequisits(): a.upload()

    #alternative with BLE
    a = ArduinoUploader("./arduino", "MC01", "/dev/ttyUSB0", NetworkProtocols.BLE)
    if a.check_prerequisits(): a.upload()


from tkinter import Frame, Label, Button, Entry
import tkinter as tk

class AddWiFiProfileView(Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.grid_columnconfigure(0, weight=0)
        self.grid_columnconfigure(1, weight=1)

        #headline
        self.headline = Label(self, text="Add New WiFi Profile")
        self.headline.grid(row=0, column=0, columnspan=2, padx=10, pady=10, sticky="w")

        #return button
        self.return_btn = Button(self, text="Return")
        self.return_btn.grid(row=0, column=1, padx=10, pady=10, sticky="e")

        #wifi key text box and entry
        self.wifi_key_label = Label(self, text="WiFi Identity:")
        self.wifi_key_label.grid(row=1, column=0, padx=10, pady=5, sticky="w")
        self.wifi_key_entry = Entry(self)
        self.wifi_key_entry.grid(row=1, column=1, padx=10, pady=5, sticky="we")

        #SSID text box and entry
        self.ssid_label = Label(self, text="SSID:")
        self.ssid_label.grid(row=2, column=0, padx=10, pady=5, sticky="w")
        self.ssid_entry = Entry(self)
        self.ssid_entry.grid(row=2, column=1, padx=10, pady=5, sticky="we")

        #Password text box and entry
        self.password_label = Label(self, text="Password:")
        self.password_label.grid(row=3, column=0, padx=10, pady=5, sticky="w")
        self.password_entry = Entry(self)
        self.password_entry.grid(row=3, column=1, padx=10, pady=5, sticky="we")

        #Client IP text box and entry
        #TODO: fill automatically
        self.client_ip_label = Label(self, text="Client IP:")
        self.client_ip_label.grid(row=4, column=0, padx=10, pady=5, sticky="w")
        self.client_ip_entry = Entry(self)
        self.client_ip_entry.grid(row=4, column=1, padx=10, pady=5, sticky="we")

        #Client Port text box and entry
        #TODO: fill automatically
        self.client_port_label = Label(self, text="Client Port:")
        self.client_port_label.grid(row=5, column=0, padx=10, pady=5, sticky="w")
        self.client_port_entry = Entry(self)
        self.client_port_entry.grid(row=5, column=1, padx=10, pady=5, sticky="we")

        #Create WiFi Profile button
        self.add_wifi_profile_btn = Button(self, text="Create")
        self.add_wifi_profile_btn.grid(row=6, column=0, padx=10, pady=10, sticky="w")

        #label for success, fail, inprogress, idle
        self.status_text = [("", "black"), ("Fail", "red"), ("Success", "green"), ("In Progress", "black")]
        i = 0
        self.status_label = Label(self, text=self.status_text[i][0], fg=self.status_text[i][1])
        self.status_label.grid(row=6, column=1, padx=10, pady=10, sticky="ew")

if __name__ == "__main__":
    root = tk.Tk()
    root.title("Add New WiFi Profile")
    root.geometry("400x500")
    add_wifi_profile_view = AddWiFiProfileView(root)
    add_wifi_profile_view.pack(fill="both", expand=True)
    root.mainloop()
